
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : menu.scm
;; DESCRIPTION : routines for generating menus
;; COPYRIGHT   : (C) 2002  Joris van der Hoeven, David Allouche
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu utilities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-empty) (widget-harray '() -1))
(define (make-menu-error . args)
  (apply tm-display-error args)
  (widget-text "Error" #f ""))

(define (pixel) 256) ; TODO: length arithmetic

(define (widget-empty) (widget-glue #f #f 0 0))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Automatic check marks
;; Should be moved to corresponding routines later
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (set-check-mark! proc mark pred)
  (set-procedure-property! proc 'check-mark (list mark pred)))
(define (check-mark proc)
  (procedure-property proc 'check-mark))

(set-check-mark! make-line-with "o" test-env?)
(set-check-mark! make-with "o" test-env?)
(set-check-mark! init-env "*" test-init?)
(set-check-mark! init-default "*" test-default?)

(define (test-shrinking-factor? n) (= (get-shrinking-factor) n))
(set-check-mark! toggle-full-screen-mode "v" full-screen?)
(set-check-mark! toggle-visible-header "v" visible-header?)
(set-check-mark! toggle-visible-icon-bar "v" visible-icon-bar?)
(set-check-mark! toggle-visible-footer "v" visible-footer?)
(set-check-mark! set-shrinking-factor "*" test-shrinking-factor?)

(define (test-preamble?) (string=? (get-env "preamble") "true"))
(define (test-page-medium? s) (string=? (get-env "page medium") s))
(define (test-page-type? s) (string=? (get-env "page type") s))
(define (test-page-orientation? s) (string=? (get-env "page orientation") s))
(define (visible-header-and-footer?)
  (equal? "true" (get-env "show header and footer")))
(define (as-on-paper?)
  (and (visible-header-and-footer?)
       (length-zero? (get-env "reduction page left margin"))
       (length-zero? (get-env "reduction page right margin"))
       (length-zero? (get-env "reduction page top margin"))
       (length-zero? (get-env "reduction page bottom margin"))))
(set-check-mark! toggle-preamble "v" test-preamble?)
(set-check-mark! set-page-medium "*" test-page-medium?)
(set-check-mark! set-page-type "*" test-page-type?)
(set-check-mark! set-page-orientation "*" test-page-orientation?)
(set-check-mark! toggle-visible-header-and-footer
		 "v" visible-header-and-footer?)
(set-check-mark! init-as-on-paper "o" as-on-paper?)
; FIXME: check init-as-on-paper; should become a persistant property

(define (selection-test-import? s) (string=? s (selection-get-import)))
(define (selection-test-export? s) (string=? s (selection-get-export)))
(define (session-test-math-input? b?) (if (session-math-input?) b? (not b?)))
(set-check-mark! selection-set-import "*" selection-test-import?)
(set-check-mark! selection-set-export "*" selection-test-export?)
(set-check-mark! session-use-math-input "*" session-test-math-input?)
(set-check-mark! toggle-insertion-position "v" test-insertion-position?)

(define (test-cell-mode? s) (string=? (get-cell-mode) s))
(define (table-test-halign? s) (string=? (table-get-halign) s))
(define (table-test-valign? s) (string=? (table-get-valign) s))
(define (cell-test-halign? s) (string=? (cell-get-halign) s))
(define (cell-test-valign? s) (string=? (cell-get-valign) s))
(define (cell-test-hmode? s) (string=? (cell-get-hmode) s))
(define (cell-test-vmode? s) (string=? (cell-get-vmode) s))
(define (cell-test-background? s) (string=? (cell-get-background) s))
(define (cell-test-vcorrect? s) (string=? (cell-get-vcorrect) s))
(define (cell-test-hyphen? s) (string=? (cell-get-hyphen) s))
(set-check-mark! set-cell-mode "*" test-cell-mode?)
(set-check-mark! table-set-halign "*" table-test-halign?)
(set-check-mark! table-set-valign "*" table-test-valign?)
(set-check-mark! toggle-table-hyphen "v" table-hyphen?)
(set-check-mark! cell-set-halign "o" cell-test-halign?)
(set-check-mark! cell-set-valign "o" cell-test-valign?)
(set-check-mark! cell-set-hmode "o" cell-test-hmode?)
(set-check-mark! cell-set-vmode "o" cell-test-vmode?)
(set-check-mark! cell-set-background "o" cell-test-background?)
(set-check-mark! cell-set-vcorrect "o" cell-test-vcorrect?)
(set-check-mark! cell-set-hyphen "o" cell-test-hyphen?)
(set-check-mark! cell-toggle-multi-paragraph "v" cell-multi-paragraph?)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu items
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (menu-label? p)
  "Does p match the syntactic structure of a menu label definition?

Label with a popup balloon. The <string> is the balloon text.
    <label> :: (balloon <label> <string>)

Label <string> drawn in black text of an arbitrary font.
    <label> :: (text <font desc> <string>)

Example default values are: family=\"roman\", class=\"mr\", series=\"medium\",
shape=\"normal\", size=10, dpi=600.
    <font desc> :: ([family [class [series [shape [size [dpi]]]]]])

Simple menu label, its display style is controlled by tt? and e? flags.
    <label> :: <string>

Pixmap menu label, the <string> is the name of the pixmap.
    <label> :: (icon <string>)"
  (match? p '(or ('balloon menu-label? string?)
		 ('text tuple? string?)
		 string?
		 ('icon string?))))

(define (make-menu-label p e? . opt)
  "p, e?, tt? = #f --> <widget>

  Assuming that (menu-label? p) is true, evaluate p to a widget.
  See the menu-label? docstring for the syntax of p"
  (let ((tt? (and (not (null? opt)) (car opt)))
	(col (color (if e? "black" "dark grey"))))
    (cond ((tuple? p 'balloon 2)        ; (balloon <label> "balloon text")
  	   (make-menu-label (cadr p) e? tt?))
  	  ((tuple? p 'text 2)		; (text <font desc> "text")
	   (widget-box (cadr p) (caddr p) col #t #t))
  	  ((string? p)			; "text"
	   (widget-menu-text p col (get-input-language) tt?))
  	  ((tuple? p 'icon 1)		; (icon "name.xpm")
  	   (widget-xpm (cadr p) #t)))))

(define (make-menu-item p e?)
  "<menu-item> :: (<label> [...] [<shortcut>] [<check>] <action>)"
  
  (define (make-button e? check label short command)
    (widget-command-button-3
     (if (string=? check "")
	 (widget-empty)
	 (widget-box '()
		     (cadr (assoc check '(("v" "<checked>")
					  ("o" "<circ>")
					  ("*" "<bullet>"))))
		     (color (if e? "black" "dark grey"))
		     #t #f))
     (make-menu-label label e?)
     (if (string=? short "")
	 (widget-empty)
	 (make-menu-label short e? #t))
     command
     e? #f))

  (define (make-shortcut label action opt-key)
    (cond (opt-key opt-key)
	  ((pair? label) "")
	  (else (get-inverse-binding action))))


  (define (make-check opt-check action)
    (or opt-check
  	(and (pair? action)
  	     (let ((proc (eval (car action))))
	       (and (procedure? proc)
  		    (let ((prop (procedure-property proc 'check-mark)))
  		      (and prop
			   (list (car prop)
				 (cons (cadr prop) (cdr action))))))))))

  (define (eval-check opt-check)
    (if (and opt-check
	     (eval-bool-or-false (cadr opt-check)))
	(car opt-check)
	""))
    
  (define (make-label label dots?)
    (if (pair? label) label
  	(string-append label (if dots? "..." ""))))

;   (set-trace-level! make-button)
;   (set-trace-level! make-shortcut)
;   (set-trace-level! make-check)
;   (set-trace-level! eval-check)
;   (set-trace-level! make-label)

    (let ((action (cAr p))
	  (label  (car p))
	  (error? #f)
	  (opt-key #f)
	  (opt-dots? #f)
	  (opt-check #f))
      (for-each
        (lambda (x)
	  (if (not error?)
	      (cond ((match? x 'menu-label?) (set! opt-key x))
		    ((match? x '...) (set! opt-dots? #t))
		    ((match? x '('check *)) (set! opt-check (cdr x)))
		    (else (set! error? #t)))))
	(cDdr p))
      (if error?
	(make-menu-error "invalid menu item attribute in " (object->string p))
  	(make-button e?
		     (eval-check (make-check opt-check action))
		     (make-label label opt-dots?)
  		     (make-shortcut label action opt-key)
		     (server-command (if e?
					 (wrap-catch-list action)
					 'noop))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu bodies
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-bad-format p e?)
  (make-menu-error "menu has bad format in " (object->string p)))

;; ENTRY POINT
(define (make-menu-widget p e?)
  "Entry point for C++ code."
  ((wrap-catch make-menu-widget-internal) p e?))

(define (make-menu-widget-internal p e?)
  "Generates menubar and menus."

  (define (make-vmenu p e?) (widget-vmenu (make-menu-items (cdr p) e?)))

  (define (make-linked-menu p e?)
    (apply-with (safe-eval-list (cadr p))
		(lambda (error? lst)
		  (if (not error?)
		      (make-menu-widget-internal lst e?)
		      (make-menu-error
		       "error evaluating link: " (object->string (cadr p))
		       ", result: " (object->string lst))))))

  (define (make-hmenu p e?) (widget-harray (make-menu-items p e?) -1))
    
;   (set-trace-point! make-vmenu      "case: vmenu")
;   (set-trace-point! make-linked-menu  "case: link")
;   (set-trace-point! make-menu-items "case: hmenu")
  
  (cond ((not (tuple? p)) (make-menu-bad-format p e?))
	((tuple? p 'ver-menu) (make-vmenu p e?))
	((tuple? p 'link 1)   (make-linked-menu p e?))
	(else (make-hmenu p e?))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu items
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-items p e?)
  (filter (flatten (make-menu-items* p e?))
	  (lambda (x) (not (null? x)))))

(define (make-menu-items* p e?)
  "Generate menu items.
  <menu-entry> :: <hsep> | <vsep> | <group> | <item-list> | <if-list>
                  | <when-list> | <item-link>"

  (define (make-hsep)
    (widget-separator (* 2 (pixel)) (* 2 (pixel)) #f))

  (define (make-vsep)
    (widget-separator (* 2 (pixel)) (* 2 (pixel)) #t))

  (define (make-group s)
    "<group> :: (group <string>)"
    (widget-command-button-3
     (widget-empty) (make-menu-label s #f) (widget-empty)
     (server-command 'noop) #f #t))

  (define (make-item-list e? p)
    "<item-list> :: ([<menu-entry> ...])"
    (map (lambda (pp) (make-menu-items* pp e?)) p))

  (define (make-items-if e? tag pred? . items)
    "<if-list> :: (if <predicate> . <item-list>)"
    (if (eval-bool-or-false pred?) (make-item-list e? items) '()))

  (define (make-items-when e? tag pred? . items)
    "<when-list> :: (when <predicate> . <item-list>)"
    (make-item-list (and e? (eval-bool-or-false pred?)) items))

  (define (make-submenu e? tag label . items)
    "<submenu> :: (<gravity> <menu-label> [<menu-entry> ...])
  <gravity> :: => | ->"
    (let ((button
	   ((cond ((equal? tag '=>) widget-pulldown-button-lazy)
		  ((equal? tag '->) widget-pullright-button-lazy))
	    (make-menu-label label e?)
	    (widget-promise (list 'make-menu-widget
				  (list 'quote (cons 'ver-menu items)) e?)))))
      (if (tuple? label 'balloon 2)
	  (widget-balloon button
			  (widget-text (caddr label) #t (get-input-language)))
	  button)))

  (define (make-tile-menu e? tag width . items)
    "<tile-menu> :: (tile <integer> [<menu-entry> ...])"
    (widget-tile (make-menu-items items e?) width))

  (define (make-linked-menu-items e? tag expr)
    "<item-link> :: (link <expr yielding <menu-entry>>)"
    (apply-with (safe-eval-list expr)
		(lambda (error? r)
		  (if (not error?)
		      (make-menu-items* r e?)
		      (make-menu-error
		       "error evaluating link: " (object->string expr))))))

  (define (make-menu-entry e? p)
    "<menu-entry> :: <menu-item>"
    (let ((button (make-menu-item p e?))
	  (label (car p)))
      (if (tuple? label 'balloon 2)
	  (widget-balloon button
 			  (widget-text (caddr label) #t (get-input-language)))
	  button)))

  (define (make-symbol-button e? symstring opt-symobj)
    (widget-command-button-1
     (widget-box '() symstring
		 (color (if e? "black" "dark grey"))
		 #t #f)
     (server-command (if opt-symobj
  			 (list 'insert-object (list 'quote opt-symobj))
  			 (list 'insert-string symstring)))
     #f))

  (define (make-menu-symbol e? tag symstring . opt)
    "<menu-symbol> :: (symbol <symbol-string> [<symbol-object>] [<shorcut>])
<symbol-object> :: (<symbol-type> <symbol-name>)
<symbol-type> :: left | right | mid | big
<symbol-name> :: <string>"
    (let ((error? #f)
	  (opt-shortcut #f)
	  (opt-symobj #f))
      (for-each
        (lambda (x)
	  (if (not error?)
	      (cond ((match? x 'pair?) (set! opt-symobj x))
		    ((match? x 'string?) (set! opt-shortcut x))
		    (else (set! error? #t)))))
	opt)
      (if error? (make-menu-error
		  "invalid symbol attribute in " (list tag symstring . opt))
	  (let ((sh (if opt-shortcut
			opt-shortcut
			(get-inverse-binding (string->symbol symstring)))))
	    (if (string=? sh "")
		(make-symbol-button e? symstring opt-symobj)
		(widget-balloon
  		 (make-symbol-button e? symstring opt-symobj)
		 (make-menu-label (string-append "Keyboard equivalent: " sh)
				  e?)))))))

;   (set-trace-point! make-hsep          "case: ---")
;   (set-trace-point! make-vsep          "case: |")
;   (set-trace-point! make-group         "case: group")
;   (set-trace-point! make-item-list     "case: list")
;   (set-trace-point! make-items-if      "case: if")
;   (set-trace-point! make-items-when    "case: when")
;   (set-trace-point! make-submenu       "case: =>/->")
;   (set-trace-point! make-tile-menu     "case: tile")
;   (set-trace-point! make-linked-menu-items "case: link")
;   (set-trace-point! make-menu-entry    "case: item")
;   (set-trace-point! make-menu-symbol   "case: symbol")

  (cond ((match? p '---) (make-hsep))
	((match? p '|) (make-vsep))
	((match? p '(group string?)) (make-group (cadr p)))
	((match? p '(menu-label? 1 *)) (make-menu-entry e? p))
	((match? p '(or () (tuple? *))) (make-item-list e? p))
	((match? p '('if 1 *)) (apply make-items-if (cons e? p)))
	((match? p '('when 1 *)) (apply make-items-when (cons e? p)))
	((match? p '('-> menu-label? *)) (apply make-submenu (cons e? p)))
	((match? p '('=> menu-label? *)) (apply make-submenu (cons e? p)))
	((match? p '('tile integer? *)) (apply make-tile-menu (cons e? p)))
	((match? p '('link 1)) (apply make-linked-menu-items (cons e? p)))
	((match? p '('symbol string? *)) (apply make-menu-symbol (cons e? p)))
	((match? p '(*)) (make-item-list e? p))
	(else (make-menu-bad-format p e?))))

; (set-trace-level! make-menu-widget)
; (set-trace-level! make-menu-item)
; (set-trace-level! make-menu-items*)
; (set-trace-level! make-menu-items)
