
/******************************************************************************
* MODULE     : tm_config.gen.cc
* DESCRIPTION: Configuration routines for TeXmacs server
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "tm_config.hh"
#include "analyze.hh"

extern bool eval_scheme_bool (string s, bool& b);

#define rewrite_find_key_binding(s) \
  find_key_binding (apply_wildcards (s, post_kbd_wildcards))

/******************************************************************************
* Constructor and destructor
******************************************************************************/

tm_config_rep::tm_config_rep ():
  in_lan ("english"),
  out_lan ("english"),
  var_suffix (" *"),
  unvar_suffix (" C-*"),
  latex_cmds (tree ("undefined latex command")),
  latex_help (""),
  keymap (""),
  invmap (""),
  shorthands (""),
  help_message ("") {}

tm_config_rep::~tm_config_rep () {}

/******************************************************************************
* Set input and output language
******************************************************************************/

void
tm_config_rep::set_input_language (string s) {
  in_lan= s;
}

void
tm_config_rep::set_output_language (string s) {
  out_lan= s;
  get_display()->set_output_language (s);
}

string
tm_config_rep::get_input_language () {
  return in_lan;
}

string
tm_config_rep::get_output_language () {
  return out_lan;
}

/******************************************************************************
* Setup TeXmacs fonts
******************************************************************************/

void
tm_config_rep::set_font_rules (scheme_tree rules) {
  int i, n= arity (rules);
  for (i=0; i<n; i++)
    if (arity (rules [i]) == 2) {
      tree l= (tree) rules[i][0];
      tree r= (tree) rules[i][1];
      font_rule (l, r);
    }  
}

/******************************************************************************
* Latex (or user) commands
******************************************************************************/

void
tm_config_rep::set_latex_commands (scheme_tree cmds) {
  int i, n= arity (cmds);
  for (i=0; i<n; i++)
    if ((arity (cmds [i]) == 3) &&
	is_atomic (cmds[i][0]) && is_atomic (cmds[i][1]))
      {
	string  l     = unquote (cmds[i][0]->label);
	latex_help (l)= unquote (cmds[i][1]->label);
	latex_cmds (l)= cmds[i][2];
      }
}

void
tm_config_rep::set_latex_symbols (scheme_tree syms) {
  int i, n= arity (syms);
  for (i=0; i<n; i++)
    if (is_atomic (syms[i])) {
      string l= syms[i]->label;
      if ((N(l)>=2) && (l[0]=='\42') && (l[N(l)-1]=='\42'))
	l= l (1, N(l)-1);
      latex_help (l)= "insert#<" * l * ">";
      latex_cmds (l)= string_to_scheme_tree ("(insert-string \"<" * l * ">\"");
    }
}

bool
tm_config_rep::get_latex_command (string which, string& help, scheme_tree& p) {
  if (!latex_help->contains (which)) return false;
  help= latex_help [which];
  p   = latex_cmds [which];
  return true;
}

/******************************************************************************
* Wildcards for keyboard bindings
******************************************************************************/

static string
apply_wildcards (string s, hashmap<string,tree> w) {
  int len, start, end, n= N(s);
  for (len=n; len>0; len--) {
    for (start=0; start <= (n-len); start++) {
      end= start+ len;
      if ((start>0) && (s[start-1] != ' ') && (s[start-1] != '-')) continue;
      if ((end<n) && (s[end-1] != ' ') && (s[end-1] != '-')) continue;
      string ss= s (start, end);
      if (s[end-1] == ' ') ss= s (start, end-1);

      // cout << "  " << ss << " => " << w[ss] << LF;
      if (w->contains (ss)) {
	tree t= w[ss];
	string rr= t[0]->label;
	bool lflag= (t[1]->label != "") || (start == 0);
	bool rflag= (t[2]->label != "") || (end == n);
	if (lflag && rflag) {
	  if ((end<n) && (rr != "") && (rr[N(rr)-1] != '-')) rr= rr * " ";
	  string r= s (0, start) * rr * s (end, n);
	  return apply_wildcards (r, w);
	}
      }      
    }
  }
  return s;
}

void
tm_config_rep::insert_kbd_wildcard (
  string key, string im, bool post, bool l, bool r)
{
  tree t= tuple (im,
		 l? string ("*"): string (""),
		 r? string ("*"): string (""));
  if (post) post_kbd_wildcards (key)= t;
  else pre_kbd_wildcards (key)= t;
}

/******************************************************************************
* Manipulation of hashmaps with keybindings and their inverses
******************************************************************************/

void
tm_config_rep::insert_key_binding (string pred, string key, scheme_tree im) {
  // cout << "Binding '" << key << "' when " << pred << " to " << im << "\n";
  tree com= im[0];
  if (!keymap->contains (key)) keymap (key)= tree (TUPLE);
  if (!invmap->contains (com)) invmap (com)= tree (TUPLE);
  delete_key_binding (pred, key);
  keymap (key)= join (tuple (tuple (pred, im )), keymap [key]);
  invmap (com)= join (tuple (tuple (pred, key)), invmap [com]);
}

void
tm_config_rep::delete_key_binding (string key) {
  // cout << "Deleting binding '" << key << "'\n";
  tree t= keymap [key];
  int i, n= N(t);
  for (i=0; i<n; i++)
    delete_inv_binding (tuple (t[i][0], key), t[i][1]);
  keymap->reset (key);
}

void
tm_config_rep::delete_key_binding (string pred, string key) {
  //cout << "Deleting binding '" << key << "' when " << pred << "\n";
  tree t= keymap [key];
  int i, n= N(t);
  for (i=0; i<n; i++)
    if (t[i][0] == pred) {
      //tree map= keymap [key];
      //keymap (key)= join (map (0, i), map (i+1, n));
      delete_inv_binding (tuple (pred, key), t[i][1]);
      keymap (key)= join (t (0, i), t (i+1, n));
      break;
    }
}

void
tm_config_rep::delete_inv_binding (tree what, scheme_tree com) {
  // cout << "Deleting inverse binding '" << com << "' for " << what << "\n";
  tree t= invmap [com];
  int i, n= N(t);
  for (i=0; i<n; i++)
    if (t[i] == what) {
      invmap (com)= join (t (0, i), t (i+1, n));
      break;
    }
}

scheme_tree
tm_config_rep::find_key_binding (string key) {
  // cout << "Find binding '" << key << "'\n";
  tree t= keymap [key];
  int i, n= N(t);
  for (i=0; i<n; i++) {
    bool found;
    (void) eval_scheme_bool ("(" * t[i][0]->label * ")", found);
    if (found) return t[i][1];
  }
  return tree (UNKNOWN);
}

string
tm_config_rep::find_inv_binding (scheme_tree com) {
  // cout << "Find inverse binding '" << com << "'\n";
  tree t= invmap [com];
  int i, n= N(t);
  for (i=0; i<n; i++) {
    bool found;
    (void) eval_scheme_bool ("(" * t[i][0]->label * ")", found);
    if (found) return t[i][1]->label;
  }
  return "";
}

/******************************************************************************
* Server keyboard mappings and shorthands
******************************************************************************/

static bool
predicate_implies (string p1, string p2) {
  if (p1 == p2) return true;
  if (p2 == "always?") return true;
  if (p2 == "in-text?")
    return
      (p1 == "in-british?") ||
      (p1 == "in-czech?") ||
      (p1 == "in-dutch?") ||
      (p1 == "in-english?") ||
      (p1 == "in-finnish?") ||
      (p1 == "in-french?") ||
      (p1 == "in-german?") ||
      (p1 == "in-hungarian?") ||
      (p1 == "in-italian?") ||
      (p1 == "in-polish?") ||
      (p1 == "in-portuguese?") ||
      (p1 == "in-romanian?") ||
      (p1 == "in-russian?") ||
      (p1 == "in-spanish?") ||
      (p1 == "in-swedish?") ||
      (p1 == "in-ukrainian?");
  return false;
}

scheme_tree
tm_config_rep::find_key_binding (string pred, string key) {
  // cout << "Find binding '" << key << "' when " << pred << "\n";
  tree t= keymap [key];
  int i, n= N(t);
  for (i=0; i<n; i++)
    if (predicate_implies (pred, t[i][0]->label))
      return t[i][1];
  return tree (UNKNOWN);
}

static string
unspace (string s) {
  int i, n= N(s);
  string r;
  for (i=0; i<n; i++)
    if (s[i] != ' ') r << s[i];
    else { i++; if (i<n) r << s[i]; }
  return r;
}

void
tm_config_rep::determine_sub_key_bindings (string pred, string s) {
  int i=0, j= 0;
  for (; i<N(s); j=i) {
    if (i!=0) i+=2;
    while ((i<N(s)) && (s[i]!=' ')) i++;
    if (i >= N(s)) break;
    if (find_key_binding (pred, s (0, i)) == tree (UNKNOWN)) {
      tree prev= find_key_binding (pred, s (0, j));
      if (is_tuple (prev) && (N(prev) == 2)) prev= prev[0];
      if ((prev == tree (UNKNOWN)) || (!is_atomic (prev))) prev= s (0, j);
      string im= prev->label * unspace (s (j, i));
      insert_key_binding (pred, s (0, i), tuple (im, ""));
    }
  }
}

void
tm_config_rep::set_keymap (scheme_tree preds, scheme_tree cmds) {
  int i, j, m= arity (cmds), n= arity (preds);
  for (j=0; j<n; j++) {
    if (!is_atomic (preds[j])) continue;
    string pred= preds[j]->label;
    if ((pred == "text") || (pred == "math") || (pred == "prog") ||
	(pred == "british") || (pred == "czech") || (pred == "dutch") ||
	(pred == "english") || (pred == "finnish") || (pred == "french") ||
	(pred == "german") || (pred == "hungarian") ||
	(pred == "italian") || (pred == "polish") ||
	(pred == "portuguese") || (pred == "romanian") ||
	(pred == "russian") || (pred == "spanish") ||
	(pred == "swedish") || (pred == "ukrainian"))
      pred= "in-" * pred * "?";

    for (i=0; i<m; i++)
      if ((arity (cmds [i]) >= 2) && is_atomic (cmds[i][0])) {
	string l= cmds[i][0]->label;
	if (is_quoted (l)) l= unquote (l);
	// string old_l= l;
	l= apply_wildcards (l, pre_kbd_wildcards);
	// cout << old_l << " -> " << l << LF;
	scheme_tree r= cmds[i][1];
	if (is_atomic (r)) {
	  if (is_quoted (r->label)) r= unquote (r->label);
	  if ((N(cmds[i]) >= 3) && is_atomic (cmds[i][2])) {
	    string help= cmds[i][2]->label;
	    if (is_quoted (help)) help= unquote (help);
	    r= tuple (r, help);
	  }
	  else r= tuple (r, "");
	}
	else {
	  if (N(cmds[i]) >= 3)
	    r= join (tuple ("begin"), cmds[i] (1, N(cmds[i])));
	  r= tuple (r, "");
	}
	determine_sub_key_bindings (pred, l);
	insert_key_binding (pred, copy (l), copy (r));
      }
  }
}

void
tm_config_rep::remove_keymap (scheme_tree preds, scheme_tree cmds) {
  int i, j, m= arity (cmds), n= arity (preds);
  for (j=0; j<n; j++) {
    if (!is_atomic (preds[j])) continue;
    string pred= preds[j]->label;
    for (i=0; i<m; i++)
      if (is_atomic (cmds[i])) {
	string l= cmds[i]->label;
	if (is_quoted (l)) l= unquote (l);
	delete_key_binding (pred, l);
      }
  }
}

void
tm_config_rep::set_variant_keys (string var, string unvar) {
  var_suffix= " " * var;
  unvar_suffix= " " * unvar;
}

void
tm_config_rep::variant_simplification (string& which) {
  if (ends (which, var_suffix)) {
    scheme_tree t= rewrite_find_key_binding (which);
    //cout << which << " => " << t << LF;
    if (!is_func (t, TUPLE, 2)) {
      while (ends (which, var_suffix))
	which= which (0, N(which) - N(var_suffix));
    }
  }
  if (ends (which, unvar_suffix)) {
    if (ends (which, var_suffix * unvar_suffix))
      which= which (0, N(which) - N(var_suffix) - N(unvar_suffix));
    else {
      which= which (0, N(which) - N(unvar_suffix));
      while (true) {
	scheme_tree t= rewrite_find_key_binding (which);
	if (!is_func (t, TUPLE, 2)) break;
	which= which * var_suffix;
      }
    }
  }
}

string
tm_config_rep::rewrite_keycomb (string s) {
  variant_simplification (s);
  return apply_wildcards (s, post_kbd_wildcards);
}

void
tm_config_rep::get_keycomb (
  string& which, int& status, scheme_tree& keym, string& shorth, string& help)
{
  //string old_which= which;
  variant_simplification (which);
  //cout << old_which << " -> " << which << LF;
  string rew= apply_wildcards (which, post_kbd_wildcards);
  //cout << which << " -> " << rew << LF;
  scheme_tree t= find_key_binding (rew);
  //cout << rew << " => " << t << LF;
  if (!is_func (t, TUPLE, 2)) {
    status= 0;
    keym  = "";
    shorth= copy (rew);
    help  = "";
  }
  else if (!is_atomic (t[0])) {
    status= 1;
    keym  = t[0];
    shorth= copy (rew);
    help  = t[1]->label;
  }
  else {
    status= 2;
    keym  = "";
    shorth= t[0]->label;
    help  = t[1]->label;
  }
}
